@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 shepherd.datalib
 ~~~~~
@@ -7,11 +6,15 @@ HDF5 files.
 
 """
 import logging
+
 from .reader import Reader
 from .writer import Writer
-# from .ivonne import Reader as IVonneReader
 
-__version__ = "2022.5.5"
-__all__ = ["Reader", "Writer", ]
+__version__ = "2023.3.1"
+
+__all__ = [
+    "Reader",
+    "Writer",
+]
 
 logging.basicConfig(format="%(name)s %(levelname)s: %(message)s", level=logging.INFO)
