@@ -1,20 +1,24 @@
 """
 Command definitions for CLI
 """
+import logging
 import os
 import sys
-import logging
 from pathlib import Path
-from typing import NoReturn
+from typing import List
+from typing import Optional
 
 import click
-from . import Writer, Reader, __version__
+
+from . import Reader
+from . import Writer
+from . import __version__
 
 logger = logging.getLogger("SHPData.cli")
 verbose_level = 2
 
 
-def config_logger(verbose: int) -> NoReturn:
+def config_logger(verbose: int) -> None:
     # TODO: put in __init__, provide logger, ditch global var
     if verbose == 0:
         logger.setLevel(logging.ERROR)
@@ -29,7 +33,7 @@ def config_logger(verbose: int) -> NoReturn:
     logging.basicConfig(format="%(message)s")  # reduce internals
 
 
-def path_to_flist(data_path: Path) -> list[Path]:
+def path_to_flist(data_path: Path) -> List[Path]:
     """every path gets transformed to a list of paths
     - if directory: list of files inside
     - if existing file: list with 1 element
@@ -49,7 +53,7 @@ def path_to_flist(data_path: Path) -> list[Path]:
     return h5files
 
 
-@click.group(context_settings=dict(help_option_names=["-h", "--help"], obj={}))
+@click.group(context_settings={"help_option_names": ["-h", "--help"], "obj": {}})
 @click.option(
     "-v",
     "--verbose",
@@ -57,8 +61,8 @@ def path_to_flist(data_path: Path) -> list[Path]:
     default=0,
     help="4 Levels [0..3](Error, Warning, Info, Debug)",
 )
-@click.pass_context
-def cli(ctx, verbose: int):
+@click.pass_context  # TODO: is the ctx-type correct?
+def cli(ctx: click.Context, verbose: int) -> None:
     """Shepherd: Synchronized Energy Harvesting Emulator and Recorder
 
     Args:
@@ -75,9 +79,8 @@ def cli(ctx, verbose: int):
 
 @cli.command(short_help="Validates a file or directory containing shepherd-recordings")
 @click.argument("in_data", type=click.Path(exists=True, resolve_path=True))
-def validate(in_data):
-    """Validates a file or directory containing shepherd-recordings
-    """
+def validate(in_data: Path) -> None:
+    """Validates a file or directory containing shepherd-recordings"""
     files = path_to_flist(in_data)
     valid_dir = True
     for file in files:
@@ -108,9 +111,8 @@ def validate(in_data):
     type=click.STRING,
     help="Set an individual csv-separator",
 )
-def extract(in_data, ds_factor, separator):
-    """Extracts recorded IVSamples and stores it to csv
-    """
+def extract(in_data: Path, ds_factor: float, separator: str) -> None:
+    """Extracts recorded IVSamples and stores it to csv"""
     files = path_to_flist(in_data)
     if not isinstance(ds_factor, (float, int)) or ds_factor < 1:
         ds_factor = 1000
@@ -158,9 +160,8 @@ def extract(in_data, ds_factor, separator):
     type=click.STRING,
     help="Set an individual csv-separator",
 )
-def extract_meta(in_data, separator):
-    """Extracts metadata and logs from file or directory containing shepherd-recordings
-    """
+def extract_meta(in_data: Path, separator: str) -> None:
+    """Extracts metadata and logs from file or directory containing shepherd-recordings"""
     files = path_to_flist(in_data)
     for file in files:
         logger.info("Extracting metadata & logs from '%s' ...", file.name)
@@ -182,7 +183,7 @@ def extract_meta(in_data, separator):
 
 @cli.command(
     short_help="Creates an array of downsampling-files from "
-               "file or directory containing shepherd-recordings"
+    "file or directory containing shepherd-recordings"
 )
 @click.argument("in_data", type=click.Path(exists=True, resolve_path=True))
 # @click.option("--out_data", "-o", type=click.Path(resolve_path=True))
@@ -196,14 +197,16 @@ def extract_meta(in_data, separator):
 @click.option(
     "--sample-rate",
     "-r",
-    type=int,
+    type=click.INT,
     help="Alternative Input to determine a downsample-factor (Choose One)",
 )
-def downsample(in_data, ds_factor, sample_rate):
-    """Creates an array of downsampling-files from file or directory containing shepherd-recordings
-    """
+def downsample(
+    in_data: Path, ds_factor: Optional[float], sample_rate: Optional[int]
+) -> None:
+    """Creates an array of downsampling-files from file
+    or directory containing shepherd-recordings"""
     if ds_factor is None and sample_rate is not None and sample_rate >= 1:
-        ds_factor = int(Reader.samplerate_sps / sample_rate)
+        ds_factor = int(Reader.samplerate_sps_default / sample_rate)
     if isinstance(ds_factor, (float, int)) and ds_factor >= 1:
         ds_list = [ds_factor]
     else:
@@ -214,7 +217,9 @@ def downsample(in_data, ds_factor, sample_rate):
         with Reader(file, verbose=verbose_level >= 2) as shpr:
             for _factor in ds_list:
                 if shpr.ds_time.shape[0] / _factor < 1000:
-                    logger.warning("will skip downsampling for %s because resulting sample-size is too small")
+                    logger.warning(
+                        "will skip downsampling for %s because resulting sample-size is too small"
+                    )
                     break
                 ds_file = file.with_suffix(f".downsampled_x{round(_factor)}.h5")
                 if ds_file.exists():
@@ -234,12 +239,8 @@ def downsample(in_data, ds_factor, sample_rate):
                     shpr.downsample(
                         shpr.ds_time, shpw.ds_time, ds_factor=_factor, is_time=True
                     )
-                    shpr.downsample(
-                        shpr.ds_voltage, shpw.ds_voltage, ds_factor=_factor
-                    )
-                    shpr.downsample(
-                        shpr.ds_current, shpw.ds_current, ds_factor=_factor
-                    )
+                    shpr.downsample(shpr.ds_voltage, shpw.ds_voltage, ds_factor=_factor)
+                    shpr.downsample(shpr.ds_current, shpw.ds_current, ds_factor=_factor)
 
 
 @cli.command(
@@ -280,9 +281,15 @@ def downsample(in_data, ds_factor, sample_rate):
     is_flag=True,
     help="Plot all files (in directory) into one Multiplot",
 )
-def plot(in_data, start: float, end: float, width: int, height: int, multiplot: bool):
-    """ Plots IV-trace from file or directory containing shepherd-recordings
-    """
+def plot(
+    in_data: Path,
+    start: Optional[float],
+    end: Optional[float],
+    width: int,
+    height: int,
+    multiplot: bool,
+) -> None:
+    """Plots IV-trace from file or directory containing shepherd-recordings"""
     files = path_to_flist(in_data)
     multiplot = multiplot and len(files) > 1
     data = []
@@ -294,11 +301,14 @@ def plot(in_data, start: float, end: float, width: int, height: int, multiplot:
             else:
                 shpr.plot_to_file(start, end, width, height)
     if multiplot:
-        logger.info(f"Got {len(data)} datasets to plot")
+        logger.info("Got %d datasets to plot", len(data))
         mpl_path = Reader.multiplot_to_file(data, in_data, width, height)
-        logger.info("Plot generated and saved to '%s'", mpl_path.name)
+        if mpl_path:
+            logger.info("Plot generated and saved to '%s'", mpl_path.name)
+        else:
+            logger.info("Plot not generated, path was already in use.")
 
 
 if __name__ == "__main__":
-    print("This File should not be executed like this ...")
+    logger.info("This File should not be executed like this ...")
     cli()
